/*
 * $Id: g_dispatch.c,v 1.6 1998/06/18 20:57:21 jochen Exp $
 * GXSNMP -- An snmp management application
 * Copyright (C) 1998 Gregory McLean & Jochen Friedrich
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.,  59 Temple Place - Suite 330, Cambridge, MA 02139, USA.
 *
 * Implementation of a SNMP dispatcher as of RFC2271
 */

#include "g_snmp.h"
#include "config.h"

/* int hash functions needed for old glibs not containing it.
 */

#ifndef HAVE_INT_HASH
gint
g_int_equal (gconstpointer v, gconstpointer v2)
{
  return *((const gint*) v) == *((const gint*) v2);
}

guint
g_int_hash (gconstpointer v)
{
  return *(const gint*) v;
}
#endif

/* This module defines the API to the SNMP RFC layer. Requests are routed
 * to the appropriate transport (e.g. IPv4 or IPv6 or IPX) by using the
 * message processing compatible with the given PDU version (V1, V2C,
 * or V3). Applications will prefer to use the sync or async event loop
 * API presented by the g_session layer.
 */

/* RFC2271 defines some dispatcher primitives as standard SNMPv3 API.
 * These names do not match GNU conventions. RFC2272 defines what exactly
 * these primitives are supposed to do.
 *
 * It appears to me these primitives are mainly designed to describe
 * interfaces between objects. I.e. the "modules" of RFC2271 translate
 * to "objects" and the "primitives" translate to "public methods".
 * In this case there should be a global registry (IR?) which keeps track
 * of all known "modules" and loads them at init time. The dispatcher
 * "module" (which should be guaranteed to exist only once -> singleton)
 * must be constructed first, so the other "modules" can register themselves.
 *
 * statusInformation =
 * sendPdu(
 *         IN   transportDomain                 (guint)
 *         IN   transportAddress                (struct sockaddr *)
 *         IN   messageProcessingModel          INTEGER (0..2147483647)
 *         IN   securityModel                   INTEGER (0..2147483647)
 *         IN   securityName                    (gpointer)
 *         IN   securityLevel                   INTEGER (1,2,3)
 *         IN   contextEngineID                 OCTET STRING (SIZE(1..32))
 *         IN   contextName                     SnmpAdminString (SIZE(0..32))
 *         IN   pduVersion                      (guint)
 *         IN   PDU                             (gpointer)
 *         IN   expectResponse                  (gboolean)
 * )
 *
 * returnResponsePdu(
 *              messageProcessingModel          INTEGER (0..2147483647)
 *              securityModel                   INTEGER (0..2147483647)
 *              securityName                    (gpointer)
 *              securityLevel                   INTEGER (1,2,3)
 *              contextEngineID                 OCTET STRING (SIZE(1..32))
 *              contextName                     SnmpAdminString (SIZE(0..32))
 *              pduVersion                      (guint)
 *              PDU                             (gpointer)
 *              maxSizeResponseScopedPDU        INTEGER (0..2147483647)
 *              stateReference                  (gpointer)
 *              statusInformation               (gint)
 * )
 *
 * registerContextEngineID(
 * )
 *
 * unregisterContextID(
 *
 * )
 *
 * Callback Functions. These must be provided by the application:
 *
 * processPdu(
 *              messageProcessingModel          INTEGER (0..2147483647)
 *              securityModel                   INTEGER (0..2147483647)
 *              securityName                    (gpointer)
 *              securityLevel                   INTEGER (1,2,3)
 *              contextEngineID                 OCTET STRING (SIZE(1..32))
 *              contextName                     SnmpAdminString (SIZE(0..32))
 *              pduVersion                      (guint)
 *              PDU                             (gpointer)
 *              maxSizeResponseScopedPDU        INTEGER (0..2147483647)
 *              stateReference                  (gpointer)
 * )
 *
 * processResponsePdu(
 *              messageProcessingModel          INTEGER (0..2147483647)
 *              securityModel                   INTEGER (0..2147483647)
 *              securityName                    (gpointer)
 *              securityLevel                   INTEGER (1,2,3)
 *              contextEngineID                 OCTET STRING (SIZE(1..32))
 *              contextName                     SnmpAdminString (SIZE(0..32))
 *              pduVersion                      (guint)
 *              PDU                             (gpointer)
 *              statusInformation               (gint)
 *              sendPduHandle                   (gpointer)
 * )
 */

/* FIXME: Should be moved into a transport module with an init function
 *        For this it would be nice, it glib already had the event loop :)
 */

int snmpsocket;

/* static registration hash tables. */

static GHashTable *message_models  = NULL;
static GHashTable *security_models = NULL;
static GHashTable *access_models   = NULL;

int 
sendPdu(guint transportDomain, struct sockaddr *transportAddress,
        guint messageProcessingModel, guint securityModel,
        GString *securityName, int securityLevel,
        GString *contextEngineID, GString *contextName,
        guint pduVersion, GString *PDU, gboolean expectResponse)
{
  gboolean         result;
  guint            destTransportDomain;
  struct sockaddr  destTransportAddress;  
  gpointer         outgoingMessage;
  guint            outgoingMessageLength;
  guint            sendPduHandle;
  struct g_message *msg_model;

/* RFC2272 explain how to send SNMP messages in 4.1. 
 *
 * Step 1) application calls our primitive
 * Step 2) we check for a known Message Processing Model
 */

  if (!message_models)
    return FALSE;

  if (!(msg_model = g_hash_table_lookup(message_models, 
      &messageProcessingModel))) 
    return FALSE;

/* Step 3) we generate a sendPduHandle
 * FIXME: don't do it such stupid!
 */

  sendPduHandle=1;

/* Step 4) sending request to version-specific Message Processing module.
 */

  result = msg_model->prepareOutgoingMessage(transportDomain, transportAddress,
             messageProcessingModel, securityModel, securityName, 
             securityLevel, contextEngineID, contextName,
             pduVersion, PDU, expectResponse, sendPduHandle,
             &destTransportDomain, &destTransportAddress,
             &outgoingMessage, &outgoingMessageLength);

/* Step 5) if the statusInformation indicates an error, return.
 */

  if (!result) return -1;

/* Step 6) send message via indicated transport.
 */

/* FIXME: This is just here for now. Should check for the transport domain 
 */

  sendto(snmpsocket, outgoingMessage, outgoingMessageLength, 0,
    &destTransportAddress, sizeof(struct sockaddr_in));

  g_free(outgoingMessage);

/* FIXME: should check for return code... 
 */

  return sendPduHandle;
}

gboolean
returnResponsePdu(guint messageProcessingModel, guint securityModel,
                  GString securityName, int securityLevel, 
                  GString contextEngineID, GString contextName, 
                  guint pduVersion, GString PDU, int maxSizeResponseScopedPDU,
                  gpointer stateReference, int statusInformation)
{

  return FALSE;
}


gboolean
g_snmp_init()
{
  message_models  = g_hash_table_new (g_int_hash, g_int_equal);
  security_models = g_hash_table_new (g_int_hash, g_int_equal);
  access_models   = g_hash_table_new (g_int_hash, g_int_equal);

/* Init all required models by RFC2271. Any private model should be
 * initialized after calling snmp_init().
 */ 

return 
  g_message_init();
#if 0
  g_security_init() &&
  g_access_init();
#endif
}

gboolean
g_register_message(guint model_nr, struct g_message *msg)
{
  guint *ptr;

  if (g_hash_table_lookup(message_models, &model_nr)) return FALSE;
  ptr = g_malloc(sizeof(guint));
  *ptr = model_nr;
  g_hash_table_insert(message_models, ptr, msg);
  return TRUE;
}

gboolean
g_register_security(guint model_nr, struct g_security *sec)
{
  guint *ptr;

  if (g_hash_table_lookup(security_models, &model_nr)) return FALSE;
  ptr = g_malloc(sizeof(guint));
  *ptr = model_nr;
  g_hash_table_insert(security_models, ptr, sec);
  return TRUE;
}

gboolean
g_register_access(guint model_nr, struct g_access *acc)
{
  guint *ptr;

  if (g_hash_table_lookup(access_models, &model_nr)) return FALSE;
  ptr = g_malloc(sizeof(guint));
  *ptr = model_nr;
  g_hash_table_insert(access_models, ptr, acc);
  return TRUE;
}
